"""L moments."""
import numpy as np
import scipy as sp

ninf = 1e-5


class Lmoments:
    """Hosking (1990) introduced the concept of L-moments, which are quantities that can be directly interpreted as scale and shape descriptors of probability distributions The L-moments of order r, denoted by λr.

    λ1 = α0 = β0
    λ2 = α0 - 2α1 = 2β1 - β0
    λ3 = α0 - 6α1 + 6α2 = 6β2 - 6β1 + β0
    λ4 = α0 - 12α1 + 30α2 - 20α3 = 20β3 - 30β2 + 12β1 - β0
    """

    def __init__(self, data):
        self.data = data
        pass

    def Lmom(self, nmom=5):
        if nmom <= 5:
            var = self.samlmusmall(nmom)
            return var
        else:
            var = self.samlmularge(nmom)
            return var

    @staticmethod
    def comb(N, k):
        if (k > N) or (N < 0) or (k < 0):
            return 0
        val = 1
        for j in range(min(k, N - k)):
            val = (val * (N - j)) // (j + 1)
        return val

    def samlmularge(self, nmom=5):

        x = self.data
        if nmom <= 0:
            raise ValueError("Invalid number of Sample L-Moments")

        x = sorted(x)
        n = len(x)

        if n < nmom:
            raise ValueError("Insufficient length of data for specified nmoments")

        ## Calculate first order
        ## Pretty efficient, no loops
        coefl1 = 1.0 / self.comb(n, 1)
        suml1 = sum(x)
        l = [coefl1 * suml1]

        if nmom == 1:
            return l[0]

        # Setup comb table, where comb[i][x] refers to comb(x,i)
        comb = []
        for i in range(1, nmom):
            comb.append([])
            for j in range(n):
                comb[-1].append(self.comb(j, i))

        for mom in range(2, nmom + 1):
            ##        print(mom)
            coefl = 1.0 / mom * 1.0 / self.comb(n, mom)
            xtrans = []
            for i in range(0, n):
                coeftemp = []
                for j in range(0, mom):
                    coeftemp.append(1)

                for j in range(0, mom - 1):
                    coeftemp[j] = coeftemp[j] * comb[mom - j - 2][i]

                for j in range(1, mom):
                    coeftemp[j] = coeftemp[j] * comb[j - 1][n - i - 1]

                for j in range(0, mom):
                    coeftemp[j] = coeftemp[j] * self.comb(mom - 1, j)

                for j in range(0, int(0.5 * mom)):
                    coeftemp[j * 2 + 1] = -coeftemp[j * 2 + 1]
                coeftemp = sum(coeftemp)
                xtrans.append(x[i] * coeftemp)

            if mom > 2:
                l.append(coefl * sum(xtrans) / l[1])
            else:
                l.append(coefl * sum(xtrans))
        return l

    def samlmusmall(self, nmom=5):
        x = self.data

        if nmom <= 0:
            raise ValueError("Invalid number of Sample L-Moments")

        x = sorted(x)
        n = len(x)

        if n < nmom:
            raise ValueError("Insufficient length of data for specified nmoments")

        ## Pretty efficient, no loops
        coefl1 = 1.0 / self.comb(n, 1)

        suml1 = sum(x)
        l1 = coefl1 * suml1

        if nmom == 1:
            ret = l1
            return ret

        # comb terms appear elsewhere, this will decrease calc time
        # for nmom > 2, and shouldn't decrease time for nmom == 2
        # comb(x,1) = x
        # for i in range(1,n+1):
        ##        comb1.append(comb(i-1,1))
        ##        comb2.append(comb(n-i,1))
        # Can be simplifed to comb1 = range(0,n)

        comb1 = range(0, n)
        comb2 = range(n - 1, -1, -1)

        coefl2 = 0.5 * 1.0 / self.comb(n, 2)
        xtrans = []
        for i in range(0, n):
            coeftemp = comb1[i] - comb2[i]
            xtrans.append(coeftemp * x[i])

        l2 = coefl2 * sum(xtrans)

        if nmom == 2:
            ret = [l1, l2]
            return ret

        ## Calculate Third order
        # comb terms appear elsewhere, this will decrease calc time
        # for nmom > 2, and shouldn't decrease time for nmom == 2
        # comb3 = comb(i-1,2)
        # comb4 = comb3.reverse()
        comb3 = []
        comb4 = []
        for i in range(0, n):
            combtemp = self.comb(i, 2)
            comb3.append(combtemp)
            comb4.insert(0, combtemp)

        coefl3 = 1.0 / 3 * 1.0 / self.comb(n, 3)
        xtrans = []
        for i in range(0, n):
            coeftemp = comb3[i] - 2 * comb1[i] * comb2[i] + comb4[i]
            xtrans.append(coeftemp * x[i])

        l3 = coefl3 * sum(xtrans) / l2

        if nmom == 3:
            ret = [l1, l2, l3]
            return ret

        ## Calculate Fourth order
        # comb5 = comb(i-1,3)
        # comb6 = comb(n-i,3)
        comb5 = []
        comb6 = []
        for i in range(0, n):
            combtemp = self.comb(i, 3)
            comb5.append(combtemp)
            comb6.insert(0, combtemp)

        coefl4 = 1.0 / 4 * 1.0 / self.comb(n, 4)
        xtrans = []
        for i in range(0, n):
            coeftemp = (
                comb5[i] - 3 * comb3[i] * comb2[i] + 3 * comb1[i] * comb4[i] - comb6[i]
            )
            xtrans.append(coeftemp * x[i])

        l4 = coefl4 * sum(xtrans) / l2

        if nmom == 4:
            ret = [l1, l2, l3, l4]
            return ret

        ## Calculate Fifth order
        comb7 = []
        comb8 = []
        for i in range(0, n):
            combtemp = self.comb(i, 4)
            comb7.append(combtemp)
            comb8.insert(0, combtemp)

        coefl5 = 1.0 / 5 * 1.0 / self.comb(n, 5)
        xtrans = []
        for i in range(0, n):
            coeftemp = (
                comb7[i]
                - 4 * comb5[i] * comb2[i]
                + 6 * comb3[i] * comb4[i]
                - 4 * comb1[i] * comb6[i]
                + comb8[i]
            )
            xtrans.append(coeftemp * x[i])

        l5 = coefl5 * sum(xtrans) / l2

        if nmom == 5:
            ret = [l1, l2, l3, l4, l5]
            return ret

    @staticmethod
    def GEV(xmom):
        """Estimate the generalized extreme value distribution parameters using Lmoments method."""
        eps = 1e-6
        maxit = 20
        # euler constant
        EU = 0.57721566
        DL2 = np.log(2)
        DL3 = np.log(3)
        A0 = 0.28377530
        A1 = -1.21096399
        A2 = -2.50728214
        A3 = -1.13455566
        A4 = -0.07138022
        B1 = 2.06189696
        B2 = 1.31912239
        B3 = 0.25077104
        C1 = 1.59921491
        C2 = -0.48832213
        C3 = 0.01573152
        D1 = -0.64363929
        D2 = 0.08985247

        T3 = xmom[2]
        # if std <= 0 or third moment > 1
        if xmom[1] <= 0 or abs(T3) >= 1:
            raise ValueError("L-Moments Invalid")

        if T3 <= 0:
            G = (A0 + T3 * (A1 + T3 * (A2 + T3 * (A3 + T3 * A4)))) / (
                1 + T3 * (B1 + T3 * (B2 + T3 * B3))
            )
            if T3 >= -0.8:
                shape = G
                GAM = np.exp(sp.special.gammaln(1 + G))
                scale = xmom[1] * G / (GAM * (1 - 2 ** (-G)))
                loc = xmom[0] - scale * (1 - GAM) / G
                para = [shape, loc, scale]
                return para

            if T3 <= -0.97:
                G = 1 - np.log(1 + T3) / DL2

            T0 = (T3 + 3) * 0.5
            for IT in range(1, maxit):
                X2 = 2 ** (-G)
                X3 = 3 ** (-G)
                XX2 = 1 - X2
                XX3 = 1 - X3
                T = XX3 / XX2
                DERIV = (XX2 * X3 * DL3 - XX3 * X2 * DL2) / (XX2**2)
                GOLD = G
                G = G - (T - T0) / DERIV
                if abs(G - GOLD) <= eps * G:
                    shape = G
                    GAM = np.exp(sp.special.gammaln(1 + G))
                    scale = xmom[1] * G / (GAM * (1 - 2 ** (-G)))
                    loc = xmom[0] - scale * (1 - GAM) / G
                    para = [shape, loc, scale]
                    return para

            print("Iteration has not converged")

        Z = 1 - T3
        G = (-1 + Z * (C1 + Z * (C2 + Z * C3))) / (1 + Z * (D1 + Z * D2))
        if abs(G) < ninf:
            scale = xmom[1] / DL2
            loc = xmom[0] - EU * scale
            para = [0, loc, scale]
            return para
        else:
            shape = G
            GAM = np.exp(sp.special.gammaln(1 + G))
            scale = xmom[1] * G / (GAM * (1 - 2 ** (-G)))
            loc = xmom[0] - scale * (1 - GAM) / G
            para = [shape, loc, scale]
            return para

    @staticmethod
    def Gumbel(mom):
        EU = 0.577215664901532861
        if mom[1] <= 0:
            raise ValueError("L-Moments Invalid")
        else:
            para2 = mom[1] / np.log(2)
            para1 = mom[0] - EU * para2
            para = [para1, para2]
            return para



from collections import OrderedDict
from typing import Any, List, Tuple, Union  # Dict,

import matplotlib.pyplot as plt
import numpy as np

# import scipy as sp
import scipy.optimize as so
from loguru import logger
from matplotlib import gridspec
from matplotlib.figure import Figure
from numpy import ndarray
from numpy.random import randint
from scipy.stats import chisquare, genextreme, gumbel_r, ks_2samp, norm

from statista.parameters import Lmoments
from statista.tools import Tools as st

ninf = 1e-5


class PlottingPosition:
    """PlottingPosition."""

    def __init__(self):
        pass

    @staticmethod
    def weibul(data: Union[list, np.ndarray], option: int = 1) -> np.ndarray:
        """Weibul.

        Weibul method to calculate the cumulative distribution function CDF or
        return period.

        Parameters
        ----------
        data : [list/array]
            list/array of the data.
        option : [1/2]
            1 to calculate the cumulative distribution function cdf or
            2 to calculate the return period.default=1

        Returns
        -------
        CDF/T: [list]
            list of cumulative distribution function or return period.
        """
        data.sort()

        if option == 1:
            CDF = np.array([j / (len(data) + 1) for j in range(1, len(data) + 1)])
            return CDF
        else:
            CDF = [j / (len(data) + 1) for j in range(1, len(data) + 1)]
            T = np.array([1 / (1 - j) for j in CDF])
            return T

    @staticmethod
    def returnPeriod(F: Union[list, np.ndarray]) -> np.ndarray:
        """returnPeriod.

        Parameters
        ----------
        F: [list/array]
            non exceedence probability.

        Returns
        -------
        array:
           return period.
        """
        F = np.array(F)
        T = 1 / (1 - F)
        return T


class Gumbel:
    """Gumbel distribution."""

    def __init__(
        self,
        data: Union[list, np.ndarray] = None,
        loc: Union[int, float] = None,
        scale: Union[int, float] = None,
    ):
        """Gumbel.

        Parameters
        ----------
        data : [list]
            data time series.
        loc: [numeric]
            location parameter
        scale: [numeric]
            scale parameter
        """
        if isinstance(data, list) or isinstance(data, np.ndarray):
            self.data = np.array(data)
            self.data_sorted = np.sort(data)
            self.cdf_Weibul = PlottingPosition.weibul(data)
            self.KStable = 1.22 / np.sqrt(len(self.data))

        self.loc = loc
        self.scale = scale
        self.Dstatic = None
        self.KS_Pvalue = None
        self.chistatic = None
        self.chi_Pvalue = None

        pass

    def pdf(
        self,
        loc: Union[float, int],
        scale: Union[float, int],
        plot_figure: bool = False,
        figsize: tuple = (6, 5),
        xlabel: str = "Actual data",
        ylabel: str = "pdf",
        fontsize: Union[float, int] = 15,
        actualdata: Union[bool, np.ndarray] = True,
    ) -> Union[Tuple[np.ndarray, Figure, Any], np.ndarray]:
        """pdf.

        Returns the value of Gumbel's pdf with parameters loc and scale at x .

        Parameters:
        -----------
        loc : [numeric]
            location parameter of the gumbel distribution.
        scale : [numeric]
            scale parameter of the gumbel distribution.

        Returns
        -------
        pdf : [array]
            probability density function pdf.
        """
        if scale <= 0:
            raise ValueError("Scale parameter is negative")

        if isinstance(actualdata, bool):
            ts = self.data
        else:
            ts = actualdata

        z = (ts - loc) / scale
        pdf = (1.0 / scale) * (np.exp(-(z + (np.exp(-z)))))
        # gumbel_r.pdf(data, loc=loc, scale=scale)
        if plot_figure:
            Qx = np.linspace(
                float(self.data_sorted[0]), 1.5 * float(self.data_sorted[-1]), 10000
            )
            pdf_fitted = self.pdf(loc, scale, actualdata=Qx)

            fig, ax = plot.pdf(
                Qx,
                pdf_fitted,
                self.data_sorted,
                figsize=figsize,
                xlabel=xlabel,
                ylabel=ylabel,
                fontsize=fontsize,
            )
            return pdf, fig, ax
        else:
            return pdf

    def cdf(
        self,
        loc: Union[float, int],
        scale: Union[float, int],
        plot_figure: bool = False,
        figsize: tuple = (6, 5),
        xlabel: str = "data",
        ylabel: str = "cdf",
        fontsize: int = 15,
        actualdata: Union[bool, np.ndarray] = True,
    ) -> Union[Tuple[np.ndarray, Figure, Any], np.ndarray]:
        """cdf.

        cdf calculates the value of Gumbel's cdf with parameters loc and scale at x.

        parameter:
        ----------
            1- loc : [numeric]
                location parameter of the gumbel distribution.
            2- scale : [numeric]
                scale parameter of the gumbel distribution.
        """
        if scale <= 0:
            raise ValueError("Scale parameter is negative")

        if isinstance(actualdata, bool):
            ts = self.data
        else:
            ts = actualdata

        z = (ts - loc) / scale
        cdf = np.exp(-np.exp(-z))

        if plot_figure:
            Qx = np.linspace(
                float(self.data_sorted[0]), 1.5 * float(self.data_sorted[-1]), 10000
            )
            cdf_fitted = self.cdf(loc, scale, actualdata=Qx)

            cdf_Weibul = PlottingPosition.weibul(self.data_sorted)

            fig, ax = plot.cdf(
                Qx,
                cdf_fitted,
                self.data_sorted,
                cdf_Weibul,
                figsize=figsize,
                xlabel=xlabel,
                ylabel=ylabel,
                fontsize=fontsize,
            )

            return cdf, fig, ax
        else:
            return cdf
        # gumbel_r.cdf(Q, loc, scale)

    def getRP(self, loc, scale, data):
        """getRP.

            getRP calculates the return period for a list/array of values or a single value.

        Parameters
        ----------
        data:[list/array/float]
            value you want the coresponding return value for
        loc: [float]
            location parameter
        scale: [float]
            scale parameter

        Returns
        -------
        float:
            return period
        """
        if isinstance(data, list) or isinstance(data, np.ndarray):
            cdf = self.cdf(loc, scale, actualdata=data)
        else:
            cdf = gumbel_r.cdf(data, loc, scale)

        rp = 1 / (1 - cdf)

        return rp

    @staticmethod
    def ObjectiveFn(p, x):
        """ObjectiveFn.

        Link : https://stackoverflow.com/questions/23217484/how-to-find-parameters-of-gumbels-distribution-using-scipy-optimize
        :param p:
        :param x:
        :return:
        """
        threshold = p[0]
        loc = p[1]
        scale = p[2]

        x1 = x[x < threshold]
        nx2 = len(x[x >= threshold])
        # pdf with a scaled pdf
        # L1 is pdf based
        L1 = (-np.log((Gumbel.pdf(x1, loc, scale) / scale))).sum()
        # L2 is cdf based
        L2 = (-np.log(1 - Gumbel.cdf(threshold, loc, scale))) * nx2
        # print x1, nx2, L1, L2
        return L1 + L2

    def estimateParameter(
        self,
        method: str = "mle",
        ObjFunc=None,
        threshold: Union[None, float, int] = None,
        test: bool = True,
    ) -> tuple:
        """estimateParameter.

        EstimateParameter estimate the distribution parameter based on MLM
        (Maximum liklihood method), if an objective function is entered as an input

        There are two likelihood functions (L1 and L2), one for values above some
        threshold (x>=C) and one for values below (x < C), now the likeliest parameters
        are those at the max value of mutiplication between two functions max(L1*L2).

        In this case the L1 is still the product of multiplication of probability
        density function's values at xi, but the L2 is the probability that threshold
        value C will be exceeded (1-F(C)).

        Parameters
        ----------
        ObjFunc : [function]
            function to be used to get the distribution parameters.
        threshold : [numeric]
            Value you want to consider only the greater values.
        method : [string]
            'mle', 'mm', 'lmoments', optimization
        test: [bool]
            Default is True.

        Returns
        -------
        Param : [list]
            scale and location parameter of the gumbel distribution.
            [loc, scale]
        """
        # obj_func = lambda p, x: (-np.log(Gumbel.pdf(x, p[0], p[1]))).sum()
        # #first we make a simple Gumbel fit
        # Par1 = so.fmin(obj_func, [0.5,0.5], args=(np.array(data),))
        method = method.lower()
        if method not in ["mle", "mm", "lmoments", "optimization"]:
            raise ValueError(
                method + "value should be 'mle', 'mm', 'lmoments' or 'optimization'"
            )
        if method == "mle" or method == "mm":
            Param = list(gumbel_r.fit(self.data, method=method))
        elif method == "lmoments":
            LM = Lmoments(self.data)
            LMU = LM.Lmom()
            Param = Lmoments.Gumbel(LMU)
        elif method == "optimization":
            if ObjFunc is None or threshold is None:
                raise TypeError("threshold should be numeric value")
            Param = gumbel_r.fit(self.data, method=method)
            # then we use the result as starting value for your truncated Gumbel fit
            Param = so.fmin(
                ObjFunc,
                [threshold, Param[0], Param[1]],
                args=(self.data,),
                maxiter=500,
                maxfun=500,
            )
            Param = [Param[1], Param[2]]

        self.loc = Param[0]
        self.scale = Param[1]

        if test:
            self.ks()
            self.chisquare()

        return Param

    @staticmethod
    def theporeticalEstimate(
        loc: Union[float, int], scale: Union[float, int], cdf: np.ndarray
    ) -> np.ndarray:
        """theporeticalEstimate.

        TheporeticalEstimate method calculates the theoretical values based on the Gumbel distribution

        Parameters:
        -----------
        param : [list]
            location ans scale parameters of the gumbel distribution.
        cdf: [list]
            cummulative distribution function/ Non Exceedence probability.

        Return:
        -------
        theoreticalvalue : [numeric]
            Value based on the theoretical distribution
        """
        if scale <= 0:
            raise ValueError("Scale parameter is negative")

        if any(cdf) <= 0 or any(cdf) > 1:
            raise ValueError("cdf Value Invalid")

        cdf = np.array(cdf)
        Qth = loc - scale * (np.log(-np.log(cdf)))

        # the main equation form scipy
        # Qth = gumbel_r.ppf(F, loc=param_dist[0], scale=param_dist[1])
        return Qth

    def ks(self) -> tuple:
        """Kolmogorov-Smirnov (KS) test.

        The smaller the D static the more likely that the two samples are drawn from the same distribution
        IF Pvalue < signeficance level ------ reject

        returns:
        --------
        Dstatic: [numeric]
            The smaller the D static the more likely that the two samples are drawn from the same distribution
        Pvalue : [numeric]
            IF Pvalue < signeficance level ------ reject the null hypotethis
        """
        if self.loc is None or self.scale is None:
            raise ValueError(
                "Value of loc/scale parameter is unknown please use "
                "'EstimateParameter' to obtain them"
            )
        Qth = self.theporeticalEstimate(self.loc, self.scale, self.cdf_Weibul)

        test = ks_2samp(self.data, Qth)
        self.Dstatic = test.statistic
        self.KS_Pvalue = test.pvalue

        print("-----KS Test--------")
        print(f"Statistic = {test.statistic}")
        if self.Dstatic < self.KStable:
            print("Accept Hypothesis")
        else:
            print("reject Hypothesis")
        print(f"P value = {test.pvalue}")
        return test.statistic, test.pvalue

    def chisquare(self) -> tuple:
        """

        Returns
        -------

        """
        if self.loc is None or self.scale is None:
            raise ValueError(
                "Value of loc/scale parameter is unknown please use "
                "'EstimateParameter' to obtain them"
            )

        Qth = self.theporeticalEstimate(self.loc, self.scale, self.cdf_Weibul)
        try:
            test = chisquare(st.standardize(Qth), st.standardize(self.data))
            self.chistatic = test.statistic
            self.chi_Pvalue = test.pvalue
            print("-----chisquare Test-----")
            print("Statistic = " + str(test.statistic))
            print("P value = " + str(test.pvalue))
            return test.statistic, test.pvalue
        except Exception as e:
            print(e)
            # raise

    def confidenceInterval(
        self,
        loc: Union[float, int],
        scale: Union[float, int],
        F: np.ndarray,
        alpha: float = 0.1,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """confidenceInterval.

        Parameters:
        -----------
        loc : [numeric]
            location parameter of the gumbel distribution.
        scale : [numeric]
            scale parameter of the gumbel distribution.
        F : [list]
            Non Exceedence probability
        alpha : [numeric]
            alpha or SignificanceLevel is a value of the confidence interval.

        Return:
        -------
        Qupper : [list]
            upper bound coresponding to the confidence interval.
        Qlower : [list]
            lower bound coresponding to the confidence interval.
        """
        if scale <= 0:
            raise ValueError("Scale parameter is negative")

        Qth = self.theporeticalEstimate(loc, scale, F)
        Y = [-np.log(-np.log(j)) for j in F]
        StdError = [
            (scale / np.sqrt(len(self.data)))
            * np.sqrt(1.1087 + 0.5140 * j + 0.6079 * j**2)
            for j in Y
        ]
        v = norm.ppf(1 - alpha / 2)
        Qupper = np.array([Qth[j] + v * StdError[j] for j in range(len(self.data))])
        Qlower = np.array([Qth[j] - v * StdError[j] for j in range(len(self.data))])
        return Qupper, Qlower

    def probapilityPlot(
        self,
        loc: float,
        scale: float,
        F: np.ndarray,
        alpha: float = 0.1,
        fig1size: tuple = (10, 5),
        fig2size: tuple = (6, 6),
        xlabel: str = "Actual data",
        ylabel: str = "cdf",
        fontsize: int = 15,
    ) -> Tuple[List[Figure], list]:
        """probapilityPlot.

        ProbapilityPlot method calculates the theoretical values based on the Gumbel distribution
        parameters, theoretical cdf (or weibul), and calculate the confidence interval.

        Parameters
        ----------
        loc : [numeric]
            location parameter of the gumbel distribution.
        scale : [numeric]
            scale parameter of the gumbel distribution.
        F : [np.ndarray]
            theoretical cdf calculated using weibul or using the distribution cdf function.
        alpha : [float]
            value between 0 and 1.
        fig1size: [tuple]
            Default is (10, 5)
        fig2size: [tuple]
            Default is (6, 6)
        xlabel: [str]
            Default is "Actual data"
        ylabel: [str]
            Default is "cdf"
        fontsize: [float]
            Default is 15.

        Returns
        -------
        Qth : [list]
            theoretical generated values based on the theoretical cdf calculated from
            weibul or the distribution parameters.
        Qupper : [list]
            upper bound coresponding to the confidence interval.
        Qlower : [list]
            lower bound coresponding to the confidence interval.
        """
        if scale <= 0:
            raise ValueError("Scale parameter is negative")

        Qth = self.theporeticalEstimate(loc, scale, F)
        Qupper, Qlower = self.confidenceInterval(loc, scale, F, alpha)

        Qx = np.linspace(
            float(self.data_sorted[0]), 1.5 * float(self.data_sorted[-1]), 10000
        )
        pdf_fitted = self.pdf(loc, scale, actualdata=Qx)
        cdf_fitted = self.cdf(loc, scale, actualdata=Qx)

        fig, ax = plot.details(
            Qx,
            Qth,
            self.data,
            pdf_fitted,
            cdf_fitted,
            F,
            Qlower,
            Qupper,
            alpha,
            fig1size=fig1size,
            fig2size=fig2size,
            xlabel=xlabel,
            ylabel=ylabel,
            fontsize=fontsize,
        )

        return fig, ax


class GEV:
    """GEV (Genalized Extreme value statistics)"""

    data: ndarray

    def __init__(
        self,
        data: Union[list, np.ndarray] = None,
        shape: Union[int, float] = None,
        loc: Union[int, float] = None,
        scale: Union[int, float] = None,
    ):
        """GEV.

        Parameters
        ----------
        data : [list]
            data time series.
        shape
        loc
        scale
        """
        if isinstance(data, list) or isinstance(data, np.ndarray):
            self.data = np.array(data)
            self.data_sorted = np.sort(data)
            self.cdf_Weibul = PlottingPosition.weibul(data)
            self.KStable = 1.22 / np.sqrt(len(self.data))

        self.shape = shape
        self.loc = loc
        self.scale = scale
        self.Dstatic = None
        self.KS_Pvalue = None

        self.chistatic = None
        self.chi_Pvalue = None
        pass

    def pdf(
        self,
        shape: Union[float, int],
        loc: Union[float, int],
        scale: Union[float, int],
        plot_figure: bool = False,
        figsize: tuple = (6, 5),
        xlabel: str = "Actual data",
        ylabel: str = "pdf",
        fontsize: int = 15,
        actualdata: Union[bool, np.ndarray] = True,
    ) -> Union[Tuple[np.ndarray, Figure, Any], np.ndarray]:
        """pdf.

        Returns the value of GEV's pdf with parameters loc and scale at x .

        Parameters
        ----------
        shape : [numeric]
            shape parameter.
        loc : [numeric]
            location parameter.
        scale : [numeric]
            scale parameter.
        plot_figure: [bool]
            Default is False.
        figsize: [tuple]
            Default is (6, 5).
        xlabel: [str]
            Default is "Actual data".
        ylabel: [str]
            Default is "pdf".
        fontsize: [int]
            Default is 15.
        actualdata : [bool/array]
            true if you want to calculate the pdf for the actual time series, array
            if you want to calculate the pdf for a theoretical time series

        Returns
        -------
        TYPE
            DESCRIPTION.
        """
        if isinstance(actualdata, bool):
            ts = self.data_sorted
        else:
            ts = actualdata

        pdf = []
        for i in range(len(ts)):
            z = (ts[i] - loc) / scale
            if shape == 0:
                val = np.exp(-(z + np.exp(-z)))
                pdf.append((1 / scale) * val)
                continue

            y = 1 - shape * z
            if y > ninf:
                # np.log(y) = ln(y)
                # ln is the inverse of e
                lnY = (-1 / shape) * np.log(y)
                val = np.exp(-(1 - shape) * lnY - np.exp(-lnY))
                pdf.append((1 / scale) * val)
                continue

            # y = 1 + shape * z
            # if y > ninf:
            #     # np.log(y) = ln(y)
            #     # ln is the inverse of e
            #     Q = y ** (-1 / shape)
            #     val = np.power(Q,1 + shape) * np.exp(-Q)
            #     pdf.append((1 / scale) * val)
            #     continue

            if shape < 0:
                pdf.append(0)
                continue
            pdf.append(0)
            # pdf.append(1)

        if len(pdf) == 1:
            pdf = pdf[0]

        pdf = np.array(pdf)
        # genextreme.pdf(data, loc=loc, scale=scale, c=shape)
        if plot_figure:
            Qx = np.linspace(
                float(self.data_sorted[0]), 1.5 * float(self.data_sorted[-1]), 10000
            )
            pdf_fitted = self.pdf(shape, loc, scale, actualdata=Qx)

            fig, ax = plot.pdf(
                Qx,
                pdf_fitted,
                self.data_sorted,
                figsize=figsize,
                xlabel=xlabel,
                ylabel=ylabel,
                fontsize=fontsize,
            )
            return pdf, fig, ax
        else:
            return pdf

    def cdf(
        self,
        shape: Union[float, int],
        loc: Union[float, int],
        scale: Union[float, int],
        plot_figure: bool = False,
        figsize: tuple = (6, 5),
        xlabel: str = "Actual data",
        ylabel: str = "cdf",
        fontsize: int = 15,
        actualdata: Union[bool, np.ndarray] = True,
    ) -> Union[Tuple[np.ndarray, Figure, Any], np.ndarray]:
        """cdf.

        Returns the value of Gumbel's cdf with parameters loc and scale
        at x.
        """
        if scale <= 0:
            raise ValueError("Scale parameter is negative")

        if isinstance(actualdata, bool):
            ts = self.data
        else:
            ts = actualdata

        z = (ts - loc) / scale
        if shape == 0:
            # GEV is Gumbel distribution
            cdf = np.exp(-np.exp(-z))
        else:
            y = 1 - shape * z
            cdf = list()
            for i in range(0, len(y)):
                if y[i] > ninf:
                    logY = -np.log(y[i]) / shape
                    cdf.append(np.exp(-np.exp(-logY)))
                elif shape < 0:
                    cdf.append(0)
                else:
                    cdf.append(1)

        cdf = np.array(cdf)

        if plot_figure:
            Qx = np.linspace(
                float(self.data_sorted[0]), 1.5 * float(self.data_sorted[-1]), 10000
            )
            cdf_fitted = self.cdf(shape, loc, scale, actualdata=Qx)

            cdf_Weibul = PlottingPosition.weibul(self.data_sorted)

            fig, ax = plot.cdf(
                Qx,
                cdf_fitted,
                self.data_sorted,
                cdf_Weibul,
                figsize=figsize,
                xlabel=xlabel,
                ylabel=ylabel,
                fontsize=fontsize,
            )

            return cdf, fig, ax
        else:
            return cdf

        # genextreme.cdf()

    def getRP(self, shape, loc, scale, data):
        """getRP.

            getRP calculates the return period for a list/array of values or a single value.

        Parameters
        ----------
        data:[list/array/float]
            value you want the coresponding return value for
        shape: [float]
            shape parameter
        loc: [float]
            location parameter
        scale: [float]
            scale parameter

        Returns
        -------
        float:
            return period
        """
        if isinstance(data, list) or isinstance(data, np.ndarray):
            cdf = self.cdf(shape, loc, scale, actualdata=data)
        else:
            cdf = genextreme.cdf(data, shape, loc, scale)

        rp = 1 / (1 - cdf)

        return rp

    def estimateParameter(
        self,
        method: str = "mle",
        ObjFunc=None,
        threshold: Union[int, float, None] = None,
        test: bool = True,
    ) -> tuple:
        """estimateParameter.

        EstimateParameter estimate the distribution parameter based on MLM
        (Maximum liklihood method), if an objective function is entered as an input

        There are two likelihood functions (L1 and L2), one for values above some
        threshold (x>=C) and one for values below (x < C), now the likeliest parameters
        are those at the max value of mutiplication between two functions max(L1*L2).

        In this case the L1 is still the product of multiplication of probability
        density function's values at xi, but the L2 is the probability that threshold
        value C will be exceeded (1-F(C)).

        Parameters
        ----------
        ObjFunc : [function]
            function to be used to get the distribution parameters.
        threshold : [numeric]
            Value you want to consider only the greater values.
        method : [string]
            'mle', 'mm', 'lmoments', optimization
        test: bool
            Default is True

        Returns
        -------
        Param : [list]
            shape, loc, scale parameter of the gumbel distribution in that order.
        """
        # obj_func = lambda p, x: (-np.log(Gumbel.pdf(x, p[0], p[1]))).sum()
        # #first we make a simple Gumbel fit
        # Par1 = so.fmin(obj_func, [0.5,0.5], args=(np.array(data),))
        method = method.lower()
        if method not in ["mle", "mm", "lmoments", "optimization"]:
            raise ValueError(
                method + "value should be 'mle', 'mm', 'lmoments' or 'optimization'"
            )

        if method == "mle" or method == "mm":
            Param = list(genextreme.fit(self.data, method=method))
        elif method == "lmoments":
            LM = Lmoments(self.data)
            LMU = LM.Lmom()
            Param = Lmoments.GEV(LMU)
        elif method == "optimization":
            if ObjFunc is None or threshold is None:
                raise TypeError("ObjFunc and threshold should be numeric value")

            Param = genextreme.fit(self.data, method=method)
            # then we use the result as starting value for your truncated Gumbel fit
            Param = so.fmin(
                ObjFunc,
                [threshold, Param[0], Param[1], Param[2]],
                args=(self.data,),
                maxiter=500,
                maxfun=500,
            )
            Param = [Param[1], Param[2], Param[3]]

        self.shape = Param[0]
        self.loc = Param[1]
        self.scale = Param[2]

        if test:
            self.ks()
            try:
                self.chisquare()
            except ValueError:
                print("chisquare test failed")

        return Param

    @staticmethod
    def theporeticalEstimate(
        shape: Union[float, int],
        loc: Union[float, int],
        scale: Union[float, int],
        F: np.ndarray,
    ) -> np.ndarray:
        """TheporeticalEstimate.

        TheporeticalEstimate method calculates the theoretical values based on a given  non exceedence probability

        Parameters:
        -----------
        param : [list]
            location ans scale parameters of the gumbel distribution.
        F : [list]
            cummulative distribution function/ Non Exceedence probability.

        Return:
        -------
        theoreticalvalue : [numeric]
            Value based on the theoretical distribution
        """
        if scale <= 0:
            raise ValueError("Parameters Invalid")

        if any(F) < 0 or any(F) > 1:
            raise ValueError("cdf Value Invalid")

        Qth = list()
        for i in range(len(F)):
            if F[i] <= 0 or F[i] >= 1:
                if F[i] == 0 and shape < 0:
                    Qth.append(loc + scale / shape)
                elif F[i] == 1 and shape > 0:
                    Qth.append(loc + scale / shape)
                else:
                    raise ValueError(str(F[i]) + " value of cdf is Invalid")
            # F = np.array(F)
            Y = -np.log(-np.log(F[i]))
            if shape != 0:
                Y = (1 - np.exp(-1 * shape * Y)) / shape

            Qth.append(loc + scale * Y)
        Qth = np.array(Qth)
        # the main equation from scipy
        # Qth = genextreme.ppf(F, shape, loc=loc, scale=scale)
        return Qth

    def ks(self):
        """Kolmogorov-Smirnov (KS) test.

        The smaller the D static the more likely that the two samples are drawn from the same distribution
        IF Pvalue < signeficance level ------ reject

        returns:
        --------
            Dstatic: [numeric]
                The smaller the D static the more likely that the two samples are drawn from the same distribution
            Pvalue : [numeric]
                IF Pvalue < signeficance level ------ reject the null hypotethis
        """
        if not hasattr(self, "loc") or not hasattr(self, "scale"):
            raise ValueError(
                "Value of loc/scale parameter is unknown please use "
                "'EstimateParameter' to obtain them"
            )
        Qth = self.theporeticalEstimate(
            self.shape, self.loc, self.scale, self.cdf_Weibul
        )

        test = ks_2samp(self.data, Qth)
        self.Dstatic = test.statistic
        self.KS_Pvalue = test.pvalue
        print("-----KS Test--------")
        print("Statistic = " + str(test.statistic))
        if self.Dstatic < self.KStable:
            print("Accept Hypothesis")
        else:
            print("reject Hypothesis")
        print("P value = " + str(test.pvalue))

        return test.statistic, test.pvalue

    def chisquare(self):
        if not hasattr(self, "loc") or not hasattr(self, "scale"):
            raise ValueError(
                "Value of loc/scale parameter is unknown please use "
                "'EstimateParameter' to obtain them"
            )

        Qth = self.theporeticalEstimate(
            self.shape, self.loc, self.scale, self.cdf_Weibul
        )

        test = chisquare(st.standardize(Qth), st.standardize(self.data))
        self.chistatic = test.statistic
        self.chi_Pvalue = test.pvalue
        print("-----chisquare Test-----")
        print("Statistic = " + str(test.statistic))
        print("P value = " + str(test.pvalue))

        return test.statistic, test.pvalue

    def confidenceInterval(
        self,
        shape: Union[float, int],
        loc: Union[float, int],
        scale: Union[float, int],
        F: np.ndarray,
        alpha: float = 0.1,
        statfunction=np.average,
        n_samples: int = 100,
        **kargs,
    ):
        """confidenceInterval.

        Parameters:
        -----------
            1- loc : [numeric]
                location parameter of the gumbel distribution.
            2- scale : [numeric]
                scale parameter of the gumbel distribution.
            1-F : [list]
                Non Exceedence probability
            3-alpha : [numeric]
                alpha or SignificanceLevel is a value of the confidence interval.

        Return:
        -------
            Qupper : [list]
                upper bound coresponding to the confidence interval.
            Qlower : [list]
                lower bound coresponding to the confidence interval.
        """
        if scale <= 0:
            raise ValueError("Scale parameter is negative")

        Param = [shape, loc, scale]
        CI = ConfidenceInterval.BootStrap(
            self.data,
            statfunction=statfunction,
            gevfit=Param,
            F=F,
            alpha=alpha,
            n_samples=n_samples,
            **kargs,
        )
        Qlower = CI["LB"]
        Qupper = CI["UB"]

        return Qupper, Qlower

    def probapilityPlot(
        self,
        shape: Union[float, int],
        loc: Union[float, int],
        scale: Union[float, int],
        F,
        alpha=0.1,
        func=None,
        n_samples=100,
        fig1size=(10, 5),
        fig2size=(6, 6),
        xlabel="Actual data",
        ylabel="cdf",
        fontsize=15,
    ):
        """probapilityPlot.

        ProbapilityPlot method calculates the theoretical values based on the Gumbel distribution
        parameters, theoretical cdf (or weibul), and calculate the confidence interval.

        Parameters
        ----------
        loc : [numeric]
            location parameter of the GEV distribution.
        scale : [numeric]
            scale parameter of the GEV distribution.
        shape: [float, int]
            shape parameter for the GEV distribution
        F : [list]
            theoretical cdf calculated using weibul or using the distribution cdf function.
        alpha : [float]
            value between 0 and 1.
        fontsize : [numeric]
            font size of the axis labels and legend
        ylabel : [string]
            y label string
        xlabel : [string]
            x label string
        fig1size : [tuple]
            size of the pdf and cdf figure
        fig2size : [tuple]
            size of the confidence interval figure
        n_samples : [integer]
            number of points in the condidence interval calculation
        alpha : [numeric]
            alpha or SignificanceLevel is a value of the confidence interval.
        func : [function]
            function to be used in the confidence interval calculation.
        """
        if scale <= 0:
            raise ValueError("Scale parameter is negative")

        Qth = self.theporeticalEstimate(shape, loc, scale, F)
        if func is None:
            func = ConfidenceInterval.GEVfunc

        Param_dist = [shape, loc, scale]
        CI = ConfidenceInterval.BootStrap(
            self.data, statfunction=func, gevfit=Param_dist, n_samples=n_samples, F=F
        )
        Qlower = CI["LB"]
        Qupper = CI["UB"]

        Qx = np.linspace(
            float(self.data_sorted[0]), 1.5 * float(self.data_sorted[-1]), 10000
        )
        pdf_fitted = self.pdf(shape, loc, scale, actualdata=Qx)
        cdf_fitted = self.cdf(shape, loc, scale, actualdata=Qx)

        fig, ax = plot.details(
            Qx,
            Qth,
            self.data,
            pdf_fitted,
            cdf_fitted,
            F,
            Qlower,
            Qupper,
            alpha,
            fig1size=fig1size,
            fig2size=fig2size,
            xlabel=xlabel,
            ylabel=ylabel,
            fontsize=fontsize,
        )

        return fig, ax


class ConfidenceInterval:
    """ConfidenceInterval."""

    def __init__(self):
        pass

    @staticmethod
    def BSIndexes(data, n_samples=10000):
        """Given data points data, where axis 0 is considered to delineate points, return an generator for sets of bootstrap indexes.

        This can be used as a list of bootstrap indexes (with
        list(bootstrap_indexes(data))) as well.
        """
        for _ in range(n_samples):
            yield randint(data.shape[0], size=(data.shape[0],))

    def BootStrap(
        data: Union[list, np.ndarray],
        statfunction,
        alpha: float = 0.05,
        n_samples: int = 100,
        **kargs,
    ):  # ->  Dict[str, OrderedDict[str, Tuple[Any, Any]]]
        """Calculate confidence intervals using parametric bootstrap and the percentil interval method This is used to obtain confidence intervals for the estimators and the return values for several return values.

        More info about bootstrapping can be found on:
            - Efron: "An Introduction to the Bootstrap", Chapman & Hall (1993)
            - https://en.wikipedia.org/wiki/Bootstrapping_%28statistics%29

        parameters:
        -----------
        alpha : [numeric]
                alpha or SignificanceLevel is a value of the confidence interval.
        kwargs :
            gevfit : [list]
                list of the three parameters of the GEV distribution [shape, loc, scale]
            F : [list]
                non exceedence probability/ cdf
        """
        alphas = np.array([alpha / 2, 1 - alpha / 2])
        tdata = (np.array(data),)

        # We don't need to generate actual samples; that would take more memory.
        # Instead, we can generate just the indexes, and then apply the statfun
        # to those indexes.
        bootindexes = ConfidenceInterval.BSIndexes(tdata[0], n_samples)
        stat = np.array(
            [
                statfunction(*(x[indexes] for x in tdata), **kargs)
                for indexes in bootindexes
            ]
        )
        stat.sort(axis=0)

        # Percentile Interval Method
        avals = alphas
        nvals = np.round((n_samples - 1) * avals).astype("int")

        if np.any(nvals == 0) or np.any(nvals == n_samples - 1):
            logger.debug(
                "Some values used extremal samples; results are probably unstable."
            )
            # warnings.warn(
            #     "Some values used extremal samples; results are probably unstable.",
            #     InstabilityWarning,
            # )
        elif np.any(nvals < 10) or np.any(nvals >= n_samples - 10):
            logger.debug(
                "Some values used top 10 low/high samples; results may be unstable."
            )
            # warnings.warn(
            #     "Some values used top 10 low/high samples; results may be unstable.",
            #     InstabilityWarning,
            # )

        if nvals.ndim == 1:
            # All nvals are the same. Simple broadcasting
            out = stat[nvals]
        else:
            # Nvals are different for each data point. Not simple broadcasting.
            # Each set of nvals along axis 0 corresponds to the data at the same
            # point in other axes.
            out = stat[(nvals, np.indices(nvals.shape)[1:].squeeze())]

        UB = out[0, 3:]
        LB = out[1, 3:]
        params = OrderedDict()
        params["shape"] = (out[0, 0], out[1, 0])
        params["location"] = (out[0, 1], out[1, 1])
        params["scale"] = (out[0, 2], out[1, 3])

        return {"LB": LB, "UB": UB, "params": params}

    # The function to bootstrap
    @staticmethod
    def GEVfunc(data, **kwargs):

        gevfit = kwargs["gevfit"]
        F = kwargs["F"]
        shape = gevfit[0]
        loc = gevfit[1]
        scale = gevfit[2]
        # generate theoretical estimates based on a random cdf, and the dist parameters
        sample = GEV.theporeticalEstimate(shape, loc, scale, np.random.rand(len(data)))
        # get parameters based on the new generated sample
        LM = Lmoments(sample)
        mum = LM.Lmom()
        newfit = LM.GEV(mum)
        shape = newfit[0]
        loc = newfit[1]
        scale = newfit[2]
        # return period
        # T = np.arange(0.1, 999.1, 0.1) + 1
        # +1 in order not to make 1- 1/0.1 = -9
        # T = np.linspace(0.1, 999, len(data)) + 1
        # coresponding theoretical estimate to T
        # F = 1 - 1 / T
        Qth = GEV.theporeticalEstimate(shape, loc, scale, F)

        res = newfit
        res.extend(Qth)
        return tuple(res)


class plot:
    """plot."""

    def __init__(self):
        pass

    def pdf(
        Qx: np.ndarray,
        pdf_fitted,
        data_sorted: np.ndarray,
        figsize: tuple = (6, 5),
        xlabel: str = "Actual data",
        ylabel: str = "pdf",
        fontsize: int = 15,
    ):
        """pdf.

        Parameters
        ----------
        pdf_fitted
        data_sorted
        figsize
        xlabel
        ylabel
        fontsize

        Returns
        -------
        """
        fig = plt.figure(figsize=figsize)
        # gs = gridspec.GridSpec(nrows=1, ncols=2, figure=fig)
        # Plot the histogram and the fitted distribution, save it for each gauge.
        ax = fig.add_subplot()
        ax.plot(Qx, pdf_fitted, "r-")
        ax.hist(data_sorted, density=True)
        ax.set_xlabel(xlabel, fontsize=fontsize)
        ax.set_ylabel(ylabel, fontsize=fontsize)
        return fig, ax

    @staticmethod
    def cdf(
        Qx,
        cdf_fitted,
        data_sorted,
        cdf_Weibul,
        figsize=(6, 5),
        xlabel="Actual data",
        ylabel="cdf",
        fontsize=15,
    ):
        """cdf.

        Parameters
        ----------
        Qx
        cdf_fitted
        data_sorted
        cdf_Weibul
        figsize
        xlabel
        ylabel
        fontsize

        Returns
        -------
        """
        fig = plt.figure(figsize=figsize)
        ax = fig.add_subplot()
        ax.plot(Qx, cdf_fitted, "r-", label="Fitted distribution")
        ax.plot(data_sorted, cdf_Weibul, ".-", label="Weibul plotting position")
        ax.set_xlabel(xlabel, fontsize=fontsize)
        ax.set_ylabel(ylabel, fontsize=fontsize)
        plt.legend(fontsize=fontsize, framealpha=1)
        return fig, ax

    @staticmethod
    def details(
        Qx: Union[np.ndarray, list],
        Qth: Union[np.ndarray, list],
        Qact: Union[np.ndarray, list],
        pdf: Union[np.ndarray, list],
        cdf_fitted: Union[np.ndarray, list],
        F: Union[np.ndarray, list],
        Qlower: Union[np.ndarray, list],
        Qupper: Union[np.ndarray, list],
        alpha: float,
        fig1size: tuple = (10, 5),
        fig2size: tuple = (6, 6),
        xlabel: str = "Actual data",
        ylabel: str = "cdf",
        fontsize: int = 15,
    ) -> Tuple[List[Figure], List[Any]]:
        """details.

        Parameters
        ----------
        Qx
        Qth
        Qact
        pdf
        cdf_fitted
        F
        Qlower
        Qupper
        alpha
        fig1size
        fig2size
        xlabel
        ylabel
        fontsize

        Returns
        -------
        """
        fig1 = plt.figure(figsize=fig1size)
        gs = gridspec.GridSpec(nrows=1, ncols=2, figure=fig1)
        # Plot the histogram and the fitted distribution, save it for each gauge.
        ax1 = fig1.add_subplot(gs[0, 0])
        ax1.plot(Qx, pdf, "r-")
        ax1.hist(Qact, density=True)
        ax1.set_xlabel(xlabel, fontsize=fontsize)
        ax1.set_ylabel("pdf", fontsize=fontsize)

        ax2 = fig1.add_subplot(gs[0, 1])
        ax2.plot(Qx, cdf_fitted, "r-")
        Qact.sort()
        ax2.plot(Qact, F, ".-")
        ax2.set_xlabel(xlabel, fontsize=fontsize)
        ax2.set_ylabel(ylabel, fontsize=15)

        fig2 = plt.figure(figsize=fig2size)
        plt.plot(Qth, Qact, "d", color="#606060", markersize=12, label="Actual Data")
        plt.plot(Qth, Qth, "^-.", color="#3D59AB", label="Theoretical Data")

        plt.plot(
            Qth,
            Qlower,
            "*--",
            color="#DC143C",
            markersize=12,
            label="Lower limit (" + str(int((1 - alpha) * 100)) + " % CI)",
        )
        plt.plot(
            Qth,
            Qupper,
            "*--",
            color="#DC143C",
            markersize=12,
            label="Upper limit (" + str(int((1 - alpha) * 100)) + " % CI)",
        )
        plt.legend(fontsize=fontsize, framealpha=1)
        plt.xlabel("Theoretical Values", fontsize=fontsize)
        plt.ylabel("Actual Values", fontsize=fontsize)

        return [fig1, fig2], [ax1, ax2]



"""Created on Mon Mar 29 21:32:29 2021.

@author: mofarrag
"""
# import Hapi.visualizer as V
import matplotlib.pyplot as plt
import numpy as np

# Vis = V.Visualize(1)


class Sensitivity:
    """Sensitivity.

    Sensitivity class

    Methods
        1- OAT
        2- Sobol
    """

    MarkerStyleList = [
        "--o",
        ":D",
        "-.H",
        "--x",
        ":v",
        "--|",
        "-+",
        "-^",
        "--s",
        "-.*",
        "-.h",
    ]

    def __init__(
        self, parameter, LB, UB, function, positions=[], n_values=5, return_values=1
    ):
        """Sensitivity.

        plotting_to instantiate the Sensitivity class you have to provide the
        following parameters

        Parameters
        ----------
        parameter : [dataframe]
            dataframe with the index as the name of the parameters and one column
            with the name "value" contains the values of the parameters.
        LB : [list]
            lower bound of the parameter.
        UB : [list]
            upper bound of the parameter.
        function : TYPE
            DESCRIPTION.
        positions : [list], optional
            position of the parameter in the list (the beginning of the list starts
            with 0), if the Position argument is empty list the sensitivity will
            be done for all parameters. The default is [].
        n_values : [integer], optional
            number of parameter values between the bounds you want to calculate the
            metric for, if the values does not include the value if the given parameter
            it will be appended to the values. The default is 5.
        return_values : [integer], optional
            return_values equals 1 if the function resurns one value (the measured metric)
            return_values equals 2 if the function resurns two values (the measured metric,
            and any calculated values you want to check how they change by changing
            the value of the parameter). The default is 1.

        Returns
        -------
        None.
        """
        self.parameter = parameter
        self.LB = LB
        self.UB = UB

        assert (
            len(self.parameter) == len(self.LB) == len(self.UB)
        ), "Length of the boundary shoulf be of the same length as the length of the parameters"
        assert callable(
            function
        ), "function should be of type callable (function that takes arguments)"
        self.function = function

        self.NoValues = n_values
        self.return_values = return_values
        # if the Position argument is empty list the sensitivity will be done for all parameters
        if positions == []:
            self.NoPar = len(parameter)
            self.Positions = list(range(len(parameter)))
        else:
            self.NoPar = len(positions)
            self.Positions = positions

    @staticmethod
    def markerStyle(style):
        """MarkerStyle.

        Marker styles for plotting

        Parameters
        ----------
        style : TYPE
            DESCRIPTION.

        Returns
        -------
        TYPE
            DESCRIPTION.
        """
        if style > len(Sensitivity.MarkerStyleList) - 1:
            style = style % len(Sensitivity.MarkerStyleList)
        return Sensitivity.MarkerStyleList[style]

    def OAT(self, *args, **kwargs):
        """OAT.

        OAT one-at-a-time sensitivity analysis.

        Parameters
        ----------
        parameter : [dataframe]
            parameters dataframe including the parameters values in a column with
            name 'value' and the parameters name as index.
        LB : [List]
            parameters upper bounds.
        UB : [List]
            parameters lower bounds.
        function : [function]
            the function you want to run it several times.
        *args : [positional argument]
            arguments of the function with the same exact names inside the function.
        **kwargs : [keyword argument]
            keyword arguments of the function with the same exact names inside the function.

        Returns
        -------
        sen : [Dict]
            for each parameter as a key, there is a list containing 4 lists,
            1-relative parameter values, 2-metric values, 3-Real parameter values
            4- adition calculated values from the function if you choose return_values=2.
        """
        self.sen = {}

        for i in range(self.NoPar):
            k = self.Positions[i]
            if self.return_values == 1:
                self.sen[self.parameter.index[k]] = [[], [], []]
            else:
                self.sen[self.parameter.index[k]] = [[], [], [], []]
            # generate 5 random values between the high and low parameter bounds
            rand_value = np.linspace(self.LB[k], self.UB[k], self.NoValues)
            # add the value of the calibrated parameter and sort the values
            rand_value = np.sort(np.append(rand_value, self.parameter["value"][k]))
            # store the relative values of the parameters in the first list in the dict
            self.sen[self.parameter.index[k]][0] = [
                ((h) / self.parameter["value"][k]) for h in rand_value
            ]

            Randpar = self.parameter["value"].tolist()
            for j in range(len(rand_value)):
                Randpar[k] = rand_value[j]
                # args = list(args)
                # args.insert(Position,Randpar)
                if self.return_values == 1:
                    metric = self.function(Randpar, *args, **kwargs)
                else:
                    metric, CalculatedValues = self.function(Randpar, *args, **kwargs)
                    self.sen[self.parameter.index[k]][3].append(CalculatedValues)
                try:
                    # store the metric value in the second list in the dict
                    self.sen[self.parameter.index[k]][1].append(round(metric, 3))
                except TypeError:
                    message = """the Given function returns more than one value,
                    the function should return only one value for return_values=1, or
                    two values for return_values=2.
                    """
                    assert False, message
                # store the real values of the parameter in the third list in the dict
                self.sen[self.parameter.index[k]][2].append(round(rand_value[j], 4))
                print(str(k) + "-" + self.parameter.index[k] + " -" + str(j))
                print(round(metric, 3))

    def Sobol(
        self,
        real_values: bool = False,
        title: str = "",  # CalculatedValues=False,
        xlabel: str = "xlabel",
        ylabel: str = "Metric values",
        labelfontsize=12,
        plotting_from="",
        plotting_to="",
        title2: str = "",
        xlabel2: str = "xlabel2",
        ylabel2: str = "ylabel2",
        spaces=[None, None, None, None, None, None],
    ):
        """Sobol.

        Parameters
        ----------
        real_values : [bool], optional
            if you want to plot the real values in the x-axis not the relative
            values, works properly only if you are checking the sensitivity of
            one parameter as the range of parameters differes. The default is False.
        CalculatedValues : [bool], optional
            if you choose return_values=2 in the OAT method, then the function returns
            calculated values, and here you can True to plot it . The default is False.
        title : [string], optional
            DESCRIPTION. The default is ''.
        xlabel : [string], optional
            DESCRIPTION. The default is 'xlabel'.
        ylabel : [string], optional
            DESCRIPTION. The default is 'Metric values'.
        labelfontsize : [integer], optional
            DESCRIPTION. The default is 12.
        plotting_from : TYPE, optional
            the calculated values are in array type and From attribute is from
            where the plotting will start. The default is ''.
        plotting_to : TYPE, optional
            the calculated values are in array type and plotting_to attribute is from
            where the plotting will end. The default is ''.
        title2 : TYPE, optional
            DESCRIPTION. The default is ''.
        xlabel2 : TYPE, optional
            DESCRIPTION. The default is 'xlabel2'.
        ylabel2 : TYPE, optional
            DESCRIPTION. The default is 'ylabel2'.
        spaces : TYPE, optional
            DESCRIPTION. The default is [None,None,None,None,None,None].

        Returns
        -------
        """
        if self.return_values == 1:
            fig, ax = plt.subplots(ncols=1, nrows=1, figsize=(8, 6))

            for i in range(self.NoPar):
                k = self.Positions[i]
                if real_values:
                    ax.plot(
                        self.sen[self.parameter.index[k]][2],
                        self.sen[self.parameter.index[k]][1],
                        Sensitivity.markerStyle(k),
                        linewidth=3,
                        markersize=10,
                        label=self.parameter.index[k],
                    )
                else:
                    ax.plot(
                        self.sen[self.parameter.index[k]][0],
                        self.sen[self.parameter.index[k]][1],
                        Sensitivity.markerStyle(k),
                        linewidth=3,
                        markersize=10,
                        label=self.parameter.index[k],
                    )

            ax.set_title(title, fontsize=12)
            ax.set_xlabel(xlabel, fontsize=12)
            ax.set_ylabel(ylabel, fontsize=12)

            ax.tick_params(axis="both", which="major", labelsize=labelfontsize)

            ax.legend(fontsize=12)
            plt.tight_layout()
            return fig, ax
        else:  # self.return_values == 2 and CalculatedValues
            try:
                fig, (ax1, ax2) = plt.subplots(ncols=1, nrows=2, figsize=(8, 6))

                for i in range(self.NoPar):
                    # for i in range(len(self.sen[self.parameter.index[0]][0])):
                    k = self.Positions[i]
                    if real_values:
                        ax1.plot(
                            self.sen[self.parameter.index[k]][2],
                            self.sen[self.parameter.index[k]][1],
                            Sensitivity.markerStyle(k),
                            linewidth=3,
                            markersize=10,
                            label=self.parameter.index[k],
                        )
                    else:
                        ax1.plot(
                            self.sen[self.parameter.index[k]][0],
                            self.sen[self.parameter.index[k]][1],
                            Sensitivity.markerStyle(k),
                            linewidth=3,
                            markersize=10,
                            label=self.parameter.index[k],
                        )

                ax1.set_title(title, fontsize=12)
                ax1.set_xlabel(xlabel, fontsize=12)
                ax1.set_ylabel(ylabel, fontsize=12)
                ax1.tick_params(axis="both", which="major", labelsize=labelfontsize)

                ax1.legend(fontsize=12)

                for i in range(self.NoPar):
                    k = self.Positions[i]
                    # for j in range(self.n_values):
                    for j in range(len(self.sen[self.parameter.index[k]][0])):
                        if plotting_from == "":
                            plotting_from = 0
                        if plotting_to == "":
                            plotting_to = len(
                                self.sen[self.parameter.index[k]][3][j].values
                            )

                        ax2.plot(
                            self.sen[self.parameter.index[k]][3][j].values[
                                plotting_from:plotting_to
                            ],
                            label=self.sen[self.parameter.index[k]][2][j],
                        )

                # ax2.legend(fontsize=12)
                box = ax2.get_position()
                ax2.set_position([box.x0, box.y0, box.width * 0.8, box.height])
                ax2.legend(loc=6, fancybox=True, bbox_to_anchor=(1.015, 0.5))

                ax2.set_title(title2, fontsize=12)
                ax2.set_xlabel(xlabel2, fontsize=12)
                ax2.set_ylabel(ylabel2, fontsize=12)

                plt.subplots_adjust(
                    left=spaces[0],
                    bottom=spaces[1],
                    right=spaces[2],
                    top=spaces[3],
                    wspace=spaces[4],
                    hspace=spaces[5],
                )

            except ValueError:
                assert (
                    False
                ), "to plot Calculated Values you should choose return_values==2 in the sentivivity object"

            plt.tight_layout()
            return fig, (ax1, ax2)

    def ListAttributes(self):
        """Print Attributes List."""

        print("\n")
        print(
            f"Attributes List of: {repr(self.__dict__['name'])} - {self.__class__.__name__} Instance\n"
        )
        self_keys = list(self.__dict__.keys())
        self_keys.sort()
        for key in self_keys:
            if key != "name":
                print(str(key) + " : " + repr(self.__dict__[key]))

        print("\n")



from numbers import Number
from typing import Union

import numpy as np
from sklearn.metrics import r2_score


def RMSE(obs: Union[list, np.ndarray], sim: Union[list, np.ndarray]) -> float:
    """Root Mean Squared Error. Metric for the estimation of performance of the hydrological model.

    Parameters
    ----------
    obs: [ndarray]
        Measured values
    sim: [ndarray]
        Simulated values

    Returns
    -------
    error: [float]
        RMSE value
    """
    # convert Qobs & sim into arrays
    obs = np.array(obs)
    sim = np.array(sim)

    rmse = np.sqrt(np.average((np.array(obs) - np.array(sim)) ** 2), dtype=np.float64)

    return rmse


def RMSEHF(
    obs: Union[list, np.ndarray],
    sim: Union[list, np.ndarray],
    WStype: int,
    N: int,
    alpha: Union[int, float],
) -> float:
    """Weighted Root mean square Error for High flow.

    Parameters
    ----------
    obs: [list/array]
        observed flow
    sim: [list/array]
        simulated flow
    WStype:
        Weighting scheme (1,2,3,4)
    N:
        power
    alpha:
        Upper limit for low flow weight

    Returns
    -------
    error values
    """
    # input data validation
    # data type
    assert isinstance(
        WStype, int
    ), f"Weighting scheme should be an integer number between 1 and 4 and you entered {WStype}"
    assert isinstance(alpha, int) or isinstance(
        alpha, float
    ), "alpha should be a number and between 0 & 1"
    assert isinstance(N, Number), "N should be a number and between 0 & 1"
    # Input values
    assert (
        1 <= WStype <= 4
    ), f"Weighting scheme should be an integer number between 1 and 4 you have enters {WStype}"
    assert (
        N >= 0
    ), f"Weighting scheme Power should be positive number you have entered {N}"
    assert (
        0 < alpha < 1
    ), f"alpha should be float number and between 0 & 1 you have entered {alpha}"

    # convert obs & sim into arrays
    obs = np.array(obs)
    sim = np.array(sim)

    Qmax = max(obs)
    h = obs / Qmax  # rational Discharge

    if WStype == 1:
        w = h**N  # rational Discharge power N
    elif (
        WStype == 2
    ):  # -------------------------------------------------------------N is not in the equation
        w = (h / alpha) ** N
        w[h > alpha] = 1
    elif WStype == 3:
        w = np.zeros(np.size(h))  # zero for h < alpha and 1 for h > alpha
        w[h > alpha] = 1
    elif WStype == 4:
        w = np.zeros(np.size(h))  # zero for h < alpha and 1 for h > alpha
        w[h > alpha] = 1
    else:  # sigmoid function
        w = 1 / (1 + np.exp(-10 * h + 5))

    a = (obs - sim) ** 2
    b = a * w
    c = sum(b)
    error = np.sqrt(c / len(obs))

    return error


def RMSELF(
    obs: Union[list, np.ndarray],
    Qsim: Union[list, np.ndarray],
    WStype: int,
    N: int,
    alpha: Union[int, float],
) -> float:
    """Weighted Root mean square Error for low flow.

    inputs:
    ----------
    obs: [list/array]
        observed flow
    sim: [list/array]
        simulated flow
    WStype:
        Weighting scheme (1,2,3,4)
    N:
        power
    alpha:
        Upper limit for low flow weight

    Returns
    -------
    error values
    """
    # input data validation
    # data type
    assert type(WStype) == int, (
        "Weighting scheme should be an integer number between 1 and 4 and you entered "
        + str(WStype)
    )
    assert isinstance(alpha, int) or isinstance(
        alpha, float
    ), "alpha should be a number and between 0 & 1"
    assert isinstance(N, Number), "N should be a number and between 0 & 1"
    # Input values
    assert (
        1 <= WStype <= 4
    ), f"Weighting scheme should be an integer number between 1 and 4 you have enters {WStype}"
    assert (
        N >= 0
    ), f"Weighting scheme Power should be positive number you have entered {N}"
    assert (
        0 < alpha < 1
    ), f"alpha should be float number and between 0 & 1 you have entered {alpha}"

    # convert obs & sim into arrays
    obs = np.array(obs)
    Qsim = np.array(Qsim)

    Qmax = max(obs)  # rational Discharge power N
    l = (Qmax - obs) / Qmax

    if WStype == 1:
        w = l**N
    elif WStype == 2:  # ------------------------------- N is not in the equation
        #        w=1-l*((0.50 - alpha)**N)
        w = ((1 / (alpha**2)) * (1 - l) ** 2) - ((2 / alpha) * (1 - l)) + 1
        w[1 - l > alpha] = 0
    elif WStype == 3:  # the same like WStype 2
        #        w=1-l*((0.50 - alpha)**N)
        w = ((1 / (alpha**2)) * (1 - l) ** 2) - ((2 / alpha) * (1 - l)) + 1
        w[1 - l > alpha] = 0
    elif WStype == 4:
        #        w = 1-l*(0.50 - alpha)
        w = 1 - ((1 - l) / alpha)
        w[1 - l > alpha] = 0
    else:  # sigmoid function
        #        w=1/(1+np.exp(10*h-5))
        w = 1 / (1 + np.exp(-10 * l + 5))

    a = (obs - Qsim) ** 2
    b = a * w
    c = sum(b)
    error = np.sqrt(c / len(obs))

    return error


def KGE(obs: Union[list, np.ndarray], sim: Union[list, np.ndarray]):
    """(Gupta et al. 2009) have showed the limitation of using a single error function to measure the efficiency of calculated flow and showed that Nash-Sutcliff efficiency (NSE) or RMSE can be decomposed into three component correlation, variability and bias.

    Parameters
    ----------
    obs: [list/array]
        observed flow
    sim: [list/array]
        simulated flow

    Returns
    -------
    error values
    """
    # convert obs & sim into arrays
    obs = np.array(obs)
    sim = np.array(sim)

    c = np.corrcoef(obs, sim)[0][1]
    alpha = np.std(sim) / np.std(obs)
    beta = np.mean(sim) / np.mean(obs)

    kge = 1 - np.sqrt(((c - 1) ** 2) + ((alpha - 1) ** 2) + ((beta - 1) ** 2))

    return kge


def WB(obs, Qsim):
    """The mean cumulative error measures how much the model succeed to reproduce the stream flow volume correctly. This error allows error compensation from time step to another and it is not an indication on how accurate is the model in the simulated flow. the naive model of Nash-Sutcliffe (simulated flow is as accurate as average observed flow) will result in WB error equals to 100 %. (Oudin et al. 2006)

     inputs:
     ----------
    obs: [list/array]
         observed flow
     sim: [list/array]
         simulated flow

     Returns
     -------
     error values
    """
    Qobssum = np.sum(obs)
    Qsimsum = np.sum(Qsim)
    wb = 100 * (1 - np.abs(1 - (Qsimsum / Qobssum)))

    return wb


def NSE(obs: np.ndarray, sim: np.ndarray):
    """Nash-Sutcliffe efficiency. Metric for the estimation of performance of the hydrological model.

    Parameters
    ----------
    obs: [list/array]
        observed flow
    sim: [list/array]
        simulated flow

    Returns
    -------
    float:
        NSE value
    """
    # convert obs & sim into arrays
    obs = np.array(obs)
    sim = np.array(sim)

    a = sum((obs - sim) ** 2)
    b = sum((obs - np.average(obs)) ** 2)
    e = 1 - (a / b)

    return e


def NSEHF(obs: Union[list, np.ndarray], sim: Union[list, np.ndarray]):
    """NSEHF.

    Modified Nash-Sutcliffe efficiency. Metric for the estimation of performance of the
    hydrological model

    reference:
    Hundecha Y. & Bárdossy A. Modeling of the effect of land use
    changes on the runoff generation of a river basin through
    parameter regionalization of a watershed model. J Hydrol
    2004, 292, (1–4), 281–295

    Parameters
    ----------
    obs: [list/array]
        observed flow
    sim: [list/array]
        simulated flow

    Returns
    -------
    float:
        NSE value
    """
    # convert obs & sim into arrays
    obs = np.array(obs)
    sim = np.array(sim)

    a = sum(obs * (obs - sim) ** 2)
    b = sum(obs * (obs - np.average(obs)) ** 2)
    e = 1 - (a / b)

    return e


def NSELF(obs: Union[list, np.ndarray], sim: Union[list, np.ndarray]):
    """NSELF.

    Modified Nash-Sutcliffe efficiency. Metric for the estimation of performance of the
    hydrological model

    reference:
    Hundecha Y. & Bárdossy A. Modeling of the effect of land use
    changes on the runoff generation of a river basin through
    parameter regionalization of a watershed model. J Hydrol
    2004, 292, (1–4), 281–295

    Parameters
    ----------
    obs: [list/array]
        observed flow
    sim: [list/array]
        simulated flow

    Returns
    -------
    float:
        NSELF value
    """
    # convert obs & sim into arrays
    obs = np.array(np.log(obs))
    sim = np.array(np.log(sim))

    a = sum(obs * (obs - sim) ** 2)
    b = sum(obs * (obs - np.average(obs)) ** 2)
    e = 1 - (a / b)

    return e


def MBE(obs: Union[list, np.ndarray], sim: Union[list, np.ndarray]):
    """MBE (mean bias error)

    MBE = (sim - obs)/n

    Parameters
    ----------
    obs: [list/array]
        observed flow
    sim: [list/array]
        simulated flow

    Returns
    -------
    float:
        mean bias error.
    """

    return (np.array(sim) - np.array(obs)).mean()


def MAE(obs: Union[list, np.ndarray], sim: Union[list, np.ndarray]):
    """MAE (mean absolute error)

    MAE = |(obs - sim)|/n

    Parameters
    ----------
    obs: [list/array]
        observed flow
    sim: [list/array]
        simulated flow

    Returns
    -------
    float:
        mean absolute error.
    """

    return np.abs(np.array(obs) - np.array(sim)).mean()


def PearsonCorre(obs: Union[list, np.ndarray], sim: Union[list, np.ndarray]):
    """Pearson correlation coefficient r2 is independent of the magnitude of the numbers; it is sensitive to relative changes only."""
    return (np.corrcoef(np.array(obs), np.array(sim))[0][1]) ** 2


def R2(obs: Union[list, np.ndarray], sim: Union[list, np.ndarray]):
    """R2.

        the coefficient of determination measures how well the predicted
        values match (and not just follow) the observed values.
        It depends on the distance between the points and the 1:1 line
        (and not the best-fit line)
        Closer the data to the 1:1 line, higher the coefficient of determination.
        The coefficient of determination is often denoted by R². However,
        it is not the square of anything. It can range from any negative number to +1
        - R² = +1 indicates that the predictions match the observations perfectly
        - R² = 0 indicates that the predictions are as good as random guesses around
            the mean of the observed values
        - Negative R² indicates that the predictions are worse than random

    Since R² indicates the distance of points from the 1:1 line, it does depend
    on the magnitude of the numbers (unlike r² peason correlation coefficient).

    Parameters
    ----------
    obs: [list/array]
        observed flow
    sim: [list/array]
        simulated flow
    """
    return r2_score(np.array(obs), np.array(sim))



"""Created on Thu May 17 04:26:42 2018.

@author: Mostafa
"""

import numpy as np


class Tools:
    """Tools.

    Tools different statistical and interpolation tools
    """

    def __init__(self):
        pass

    @staticmethod
    def normalize(x):
        """Normalizer.

        to normalize values between 0 and 1

        Parameters
        ----------
        x : [List]
            list of values

        Returns
        -------
        normalized numbers : [List]
            list of normalized values
        """
        x = np.array(x)
        DataMax = max(x)
        DataMin = min(x)
        N = (x - DataMin) / (DataMax - DataMin)
        # [i - DataMin / (DataMax - DataMin) for i in x]
        return N

    @staticmethod
    def standardize(x):
        """Standardize.

        to standardize (make the average equals 1 and the standard deviation
        equals 0)

        Parameters
        ----------
        x: [List]
            list of values

        Returns
        -------
        Standardized values: [List]
            list of normalized values
        """
        x = np.array(x)

        mean = np.mean(x)
        std = np.std(x)
        s = (x - mean) / std
        # [i - mean / (std) for i in x]
        return s

    @staticmethod
    def rescale(OldValue, OldMin, OldMax, NewMin, NewMax):
        """Rescale.

        Rescale nethod rescales a value between two boundaries to a new value
        bewteen two other boundaries

        Parameters
        ----------
        OldValue: [float]
            value need to transformed
        OldMin: [float]
            min old value
        OldMax: [float]
            max old value
        NewMin: [float]
            min new value
        NewMax: [float]
            max new value

        Returns
        -------
        NewValue: [float]
            transformed new value
        """
        OldRange = OldMax - OldMin
        NewRange = NewMax - NewMin
        NewValue = (((OldValue - OldMin) * NewRange) / OldRange) + NewMin

        return NewValue

    @staticmethod
    def logarithmicRescale(x, min_old, max_old, min_new, max_new):
        """LogarithmicRescale.

        this function transform the value between two normal values to a logarithmic scale
        between logarithmic value of both boundaries
            np.log(base)(number) = power
            the inverse of logarithmic is base**power = number

        Parameters
        ----------
        x: [float]
            new value needed to be transformed to a logarithmic scale
        min_old: [float]
            min old value in normal scale
        max_old: [float]
            max old value in normal scale
        min_new: [float]
            min new value in normal scale
        max_new: [float]
            max_new max new value

        Returns
        -------
        Y: [int]
            integer number between new max_new and min_new boundaries
        """
        # get the boundaries of the logarithmic scale
        if min_old == 0.0:
            min_old_log = -7
        else:
            min_old_log = np.log(min_old)

        max_old_log = np.log(max_old)

        if x == 0:
            x_log = -7
        else:
            x_log = np.log(x)

        y = int(
            np.round(Tools.rescale(x_log, min_old_log, max_old_log, min_new, max_new))
        )

        return y

    @staticmethod
    def invLogarithmicRescale(x, min_old, max_old, min_new, max_new, base=np.e):
        """LogarithmicRescale.

        this function transform the value between two normal values to a logarithmic scale
        between logarithmic value of both boundaries
            np.log(base)(number) = power
            the inverse of logarithmic is base**power = number

        Parameters
        ----------
        x: [float]
            new value needed to be transformed to a logarithmic scale
        min_old: [float]
            min old value in normal scale
        max_old: [float]
            max old value in normal scale
        min_new: [float]
            min new value in normal scale
        max_new: [float]
            max_new max new value

        Returns
        -------
        Y: [int]
            integer number between new max_new and min_new boundaries
        """
        # get the boundaries of the logarithmic scale

        min_old_power = np.power(base, min_old)
        max_old_power = np.power(base, max_old)
        x_power = np.power(base, x)

        y = int(
            np.round(
                Tools.rescale(x_power, min_old_power, max_old_power, min_new, max_new)
            )
        )
        return y

    @staticmethod
    def round(number, roundto):
        return round(number / roundto) * roundto



try:
    from importlib.metadata import PackageNotFoundError  # type: ignore
    from importlib.metadata import version
except ImportError:  # pragma: no cover
    from importlib_metadata import PackageNotFoundError  # type: ignore
    from importlib_metadata import version


try:
    __version__ = version(__name__)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"

# documentation format
__author__ = "Mostafa Farrag"
__email__ = "moah.farag@gmail.com"
__docformat__ = "restructuredtext"

# Let users know if they're missing any of our hard dependencies
hard_dependencies = ()  # ("numpy", "pandas", "gdal")
missing_dependencies = []

for dependency in hard_dependencies:
    try:
        __import__(dependency)
    except ImportError as e:
        missing_dependencies.append(dependency)
        print(e)

if missing_dependencies:
    raise ImportError("Missing required dependencies {0}".format(missing_dependencies))

# import statista.distributions
# import statista.metrics
# import statista.parameters
# import statista.sensitivity
# import statista.tools

__doc__ = """
statista - statistics package
"""



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['statista']

package_data = \
{'': ['*']}

install_requires = \
['loguru>=0.6.0,<0.7.0',
 'matplotlib>=3.5.3,<4.0.0',
 'pandas>=1.4.4,<2.0.0',
 'scikit-learn>=1.1.1,<2.0.0',
 'scipy>=1.9.1,<2.0.0']

setup_kwargs = {
    'name': 'statista',
    'version': '0.1.6',
    'description': 'statistics package',
    'long_description': '[![Python Versions](https://img.shields.io/pypi/pyversions/statista.png)](https://img.shields.io/pypi/pyversions/statista)\n[![License: GPL v3](https://img.shields.io/badge/License-GPLv3-blue.svg)](https://www.gnu.org/licenses/gpl-3.0)\n[![pre-commit](https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white)](https://github.com/pre-commit/pre-commit)\n[![Language grade: Python](https://img.shields.io/lgtm/grade/python/g/MAfarrag/earth2observe.svg?logo=lgtm&logoWidth=18)](https://lgtm.com/projects/g/MAfarrag/earth2observe/context:python)\n[![Total alerts](https://img.shields.io/lgtm/alerts/g/MAfarrag/earth2observe.svg?logo=lgtm&logoWidth=18)](https://lgtm.com/projects/g/MAfarrag/earth2observe/alerts/)\n\n\n\n![GitHub last commit](https://img.shields.io/github/last-commit/MAfarrag/statista)\n![GitHub forks](https://img.shields.io/github/forks/MAfarrag/statista?style=social)\n![GitHub Repo stars](https://img.shields.io/github/stars/MAfarrag/statista?style=social)\n\n\nCurrent release info\n====================\n\n| Name | Downloads                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                         | Version | Platforms |\n| --- |---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------| --- | --- |\n| [![Conda Recipe](https://img.shields.io/badge/recipe-statista-green.svg)](https://anaconda.org/conda-forge/statista) | [![Conda Downloads](https://img.shields.io/conda/dn/conda-forge/statista.svg)](https://anaconda.org/conda-forge/statista) [![Downloads](https://pepy.tech/badge/statista)](https://pepy.tech/project/statista) [![Downloads](https://pepy.tech/badge/statista/month)](https://pepy.tech/project/statista)  [![Downloads](https://pepy.tech/badge/statista/week)](https://pepy.tech/project/statista)  ![PyPI - Downloads](https://img.shields.io/pypi/dd/statista?color=blue&style=flat-square) ![GitHub all releases](https://img.shields.io/github/downloads/MAfarrag/statista/total) ![GitHub release (latest by date)](https://img.shields.io/github/downloads/MAfarrag/statista/0.1.3/total) | [![Conda Version](https://img.shields.io/conda/vn/conda-forge/statista.svg)](https://anaconda.org/conda-forge/statista) [![PyPI version](https://badge.fury.io/py/statista.svg)](https://badge.fury.io/py/statista) [![Anaconda-Server Badge](https://anaconda.org/conda-forge/statista/badges/version.svg)](https://anaconda.org/conda-forge/statista) | [![Conda Platforms](https://img.shields.io/conda/pn/conda-forge/statista.svg)](https://anaconda.org/conda-forge/statista) [![Join the chat at https://gitter.im/Hapi-Nile/Hapi](https://badges.gitter.im/Hapi-Nile/Hapi.svg)](https://gitter.im/Hapi-Nile/Hapi?utm_source=badge&utm_medium=badge&utm_campaign=pr-badge&utm_content=badge) |\n\nstatista - Statistics package\n=====================================================================\n**statista** is a statistics package\n\nstatista\n\nMain Features\n-------------\n  -\n\n\nFuture work\n-------------\n  -\n\n\n\nInstalling statista\n===============\n\nInstalling `statista` from the `conda-forge` channel can be achieved by:\n\n```\nconda install -c conda-forge statista\n```\n\nIt is possible to list all of the versions of `statista` available on your platform with:\n\n```\nconda search statista --channel conda-forge\n```\n\n## Install from Github\nto install the last development to time you can install the library from github\n```\npip install git+https://github.com/MAfarrag/statista\n```\n\n## pip\nto install the last release you can easly use pip\n```\npip install statista==0.1.6\n```\n\nQuick start\n===========\n\n```\n  >>> import statista\n```\n\n[other code samples](https://statista.readthedocs.io/en/latest/?badge=latest)\n',
    'author': 'Mostafa Farrag',
    'author_email': 'moah.farag@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/MAfarrag/statista',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<3.12',
}


setup(**setup_kwargs)
